from typing import List, Any
import typer
import logging
import signal
from rich.logging import RichHandler
from rich.console import Console
from functools import partial

from fact_lake.config import get_logger
from fact_lake.server import Server

from os import environ as env


console = Console()
typer_app = typer.Typer()


FORMAT = "%(message)s"

log = get_logger("cli")

# * Allow client side forking
# c.f.: https://github.com/grpc/grpc/blob/master/doc/fork_support.md
env["GRPC_ENABLE_FORK_SUPPORT"] = "true"
env["GRPC_POLL_STRATEGY"] = "poll"


def sigusr1handler(server: Server, signum: int, frame: Any) -> None:
    server.print_process_status()


@typer_app.callback()
def main(
    ctx: typer.Context,
    verbose: List[bool] = typer.Option(
        [False],
        "--verbose",
        "-v",
        help="Output more details on what is happening. Use multiple times for more details.",
    ),
) -> None:
    level = "WARNING"
    if len(verbose) == 1:
        level = "INFO"
    if len(verbose) >= 2:
        level = "DEBUG"

    logging.basicConfig(
        level=level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )


@typer_app.command()
def run() -> None:
    server = Server(console=console)
    signal.signal(signal.SIGUSR1, partial(sigusr1handler, server))
    server.run()


# TODO status command

if __name__ == "__main__":
    typer_app()



import logging
from os import environ as env
from pathlib import Path
from typing import Any, Optional, Dict, Tuple, Type, TypeVar, List
from pyfactcast.client.config import ClientConfiguration, Credentials
from pydantic import BaseSettings, validator

log = logging.getLogger("fact-lake")

log_level_from_env = env.get("FACT_LAKE_LOG_LEVEL")
if log_level_from_env:
    log_level = getattr(
        logging,
        log_level_from_env.upper(),
    )  # Setting a default here is pretty defensive. Anyhow nothing lost by doing it
    log.setLevel(log_level)


def get_logger(name: str) -> logging.Logger:
    return log.getChild(name)


C = TypeVar("C", bound="AppConfiguration.Config")


class AppConfiguration(BaseSettings):
    cryptoshred_init_vector: str = ""
    decrypt: bool = True
    mongo_username: str
    mongo_password: str
    factcast_server: str = "localhost"
    factcast_user: Optional[str] = None
    factcast_password: Optional[str] = None
    factcast_root_cert_path: Optional[str] = None
    factcast_cn_overwrite: Optional[str] = None
    factcast_insecure: bool = False
    mongo_server: str = "localhost"
    mongo_port: int = 27017
    mongo_tls: bool = False
    mongo_tls_invalid_certificates: bool = False
    mongo_tls_ca_file_path: Optional[str] = None
    mongodb_retry_writes: bool = False
    whitelist: List[str] = []
    blacklist: List[str] = []

    @validator("blacklist")
    def only_black_or_white_list(
        cls, v: Any, values: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> None:
        if "whitelist" in values and values["whitelist"] and v:
            raise ValueError(
                "Whitelist and blacklist defined at the same time. Only define one."
            )

    class Config:
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls: Type[C],
            init_settings: Dict,
            env_settings: Dict,
            file_secret_settings: Dict,
        ) -> Tuple:
            return (env_settings, init_settings, file_secret_settings)


def get_configuration(
    profile: str = "default", *, config_dir: Path = Path.home().joinpath(".fact_lake")
) -> AppConfiguration:
    log.info("Getting Configuration")

    if profile:
        env_file_location = config_dir.joinpath(f"{profile}.env").absolute()
        return AppConfiguration(_env_file=env_file_location)

    return AppConfiguration()


def get_factcast_configuration(config: AppConfiguration) -> ClientConfiguration:
    credentials = None
    if config.factcast_user:
        credentials = Credentials(
            username=config.factcast_user, password=config.factcast_password
        )

    return ClientConfiguration(
        server=config.factcast_server,
        root_cert_path=config.factcast_root_cert_path,
        ssl_target_override=config.factcast_cn_overwrite,
        credentials=credentials,
        insecure=config.factcast_insecure,
    )



from typing import Tuple
import grpc
import pymongo
from pymongo.collection import Collection

import cryptoshred.backends
from cryptoshred import convenience as CS
from pyfactcast.client import sync as FS
from pyfactcast.client.entities import Fact

from fact_lake.config import get_configuration, get_factcast_configuration, get_logger

app_config = get_configuration()  # TODO make configurable
fs_config = get_factcast_configuration(app_config)

log = get_logger("playing")


key_backend = cryptoshred.backends.DynamoDbSsmBackend(
    iv_param=app_config.cryptoshred_init_vector
)

fact_store = FS.FactStore(
    FS.get_synchronous_grpc_client(get_factcast_configuration(app_config))
)


def _connect_to_mongo(namespace: str) -> Tuple[Collection, Collection]:
    mongo_client = pymongo.MongoClient(
        host=app_config.mongo_server,
        port=app_config.mongo_port,
        username=app_config.mongo_username,
        password=app_config.mongo_password,
        tls=app_config.mongo_tls,
        tlsAllowInvalidCertificates=app_config.mongo_tls_invalid_certificates,
        tlsCAFile=app_config.mongo_tls_ca_file_path,
        retryWrites=app_config.mongodb_retry_writes,
    )
    db = mongo_client.get_database("facts")
    state_collection = db["last_seen"]
    collection = db[namespace]
    return (collection, state_collection)


def project_namespace(namespace: str) -> None:

    # * https://pymongo.readthedocs.io/en/stable/faq.html#is-pymongo-fork-safe
    collection, state_collection = _connect_to_mongo(namespace=namespace)

    state = state_collection.find_one(filter={"namespace": namespace})
    last_seen = None
    if state:
        last_seen = state["last_seen"]

    try:
        with fact_store as fs:
            for item in fs.subscribe(
                subscription_specs=[
                    FS.SubscriptionSpec(
                        ns=namespace,
                    )
                ],
                after_fact=last_seen,
                continuous=True,
            ):
                log.info(f"[{namespace}] Next fact")
                fact_id = _insert_fact(item=item, collection=collection)

                state_collection.find_one_and_update(
                    filter={"namespace": namespace},
                    update={"$set": {"last_seen": fact_id}},
                    upsert=True,
                )

    except grpc._channel._MultiThreadedRendezvous:
        log.warning("GRPC error")
        project_namespace(namespace=namespace)


def _insert_fact(item: Fact, collection: Collection) -> str:
    fact_id = item.header.id
    mongo_item = item.dict()
    mongo_item["_id"] = fact_id

    res = CS.find_and_decrypt_in_dict(input=[item.payload], key_backend=key_backend)
    mongo_item["payload"] = res[0]
    collection.insert_one(mongo_item).inserted_id
    return str(fact_id)



from multiprocessing import Process
from time import sleep
from typing import List, Optional
from rich.console import Console

from pyfactcast.client import sync as FS

from fact_lake.config import get_configuration, get_factcast_configuration, get_logger
from fact_lake.projectors import project_namespace


app_config = get_configuration()  # TODO make configurable
fs_config = get_factcast_configuration(app_config)


fact_store = FS.FactStore(
    FS.get_synchronous_grpc_client(get_factcast_configuration(app_config))
)


class Server:
    def __init__(self, console: Optional[Console] = None) -> None:
        self.processes: List[Process] = []
        self.log = get_logger(self.__class__.__name__)
        if not console:
            self.console = Console()
        else:
            self.console = console

    def run(self) -> None:
        with fact_store as fs:
            namespaces = set(fs.enumerate_namespaces())

            if app_config.whitelist:
                namespaces_to_work_on = namespaces.intersection(app_config.whitelist)
            elif app_config.blacklist:
                namespaces_to_work_on = namespaces - set(app_config.blacklist)
            else:
                namespaces_to_work_on = namespaces

            self.processes = [
                Process(target=project_namespace, args=(namespace,), name=namespace)
                for namespace in namespaces_to_work_on
            ]

            for process in self.processes:
                process.start()

            self._restart_dead_namespaces()

    def _restart_dead_namespaces(self) -> None:
        while True:
            sleep(60)
            self.log.info("Checking processes for liveness.")
            for process in self.processes:
                namespace = process.name

                if not process.is_alive():
                    self.log.debug(
                        f"Found dead process for namespace {namespace}. Restarting..."
                    )
                    self.processes.remove(process)
                    replacement = Process(
                        target=project_namespace, args=(namespace,), name=namespace
                    )
                    self.processes.append(replacement)
                    replacement.start()
                    self.log.debug("Restarted.")
                else:
                    self.log.debug(f"Process for namespace {namespace} is alive.")

    def print_process_status(self) -> None:
        for process in self.processes:
            self.console.print(
                f"Process for namespace {process.name} is alive: {process.is_alive()}"
            )

    def join(self) -> None:
        for process in self.processes:
            process.join()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fact_lake']

package_data = \
{'': ['*']}

install_requires = \
['cryptoshred>=0.0.7,<0.0.8',
 'pydantic>=1.9.0,<2.0.0',
 'pyfactcast>=0.0.9,<0.0.10',
 'pymongo>=4.0.1,<5.0.0',
 'rich<11',
 'typer<0.4']

extras_require = \
{'docs': ['sphinx<4',
          'sphinx-click>=2.7,<3.0',
          'sphinx-rtd-theme>=0.5,<0.6',
          'sphinx-autodoc-typehints>=1.12,<2.0']}

entry_points = \
{'console_scripts': ['fact-lake = fact_lake.cli:typer_app']}

setup_kwargs = {
    'name': 'fact-lake',
    'version': '0.0.4',
    'description': 'A datalake that is keeping up to date with your factcast',
    'long_description': '# Fact Lake\n\nJust a project stub for now\n\nContributions are welcome. Just get in touch.\n\n## Quickstart\n\nSimply `pip install fact-lake` and get going.\n\n## Development\n\nThis project uses `poetry` for dependency management and `pre-commit` for local checks.\n',
    'author': 'Eduard Thamm',
    'author_email': 'eduard.thamm@thammit.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/friendly-facts/fact-lake',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
